@@ -1,6 +1,7 @@
 /*-
  * Copyright (c) 2002 Dag-Erling Coïdan Smørgrav
  * Copyright (c) 2008 William Pitcock
+ * Copyright (c) 2018 Thorsten Alteholz
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,21 +43,33 @@
 #include <grp.h>
 #include <assert.h>
 #include <limits.h>
+#include <err.h>
 
-#define VERSION "0.3.1"
+#define VERSION "0.4.0"
 
-#define SEARCH_ALL     0x00
-#define SEARCH_GID     0x01
-#define SEARCH_PID     0x02
-#define SEARCH_PNAME   0x04
-#define SEARCH_UID     0x08
-#define SEARCH_LISTEN  0x10
-#define SEARCH_CONN    0x20
-#define SEARCH_PORTS   0x40
+#define SEARCH_ALL		0x0000
+#define SEARCH_GID		0x0001
+#define SEARCH_PID		0x0002
+#define SEARCH_PNAME		0x0004
+#define SEARCH_UID		0x0008
+#define SEARCH_LISTEN		0x0010
+#define SEARCH_CONN		0x0020
+#define SEARCH_PORTS		0x0040
+#define SEARCH_IPV4_ONLY	0x0080
+#define SEARCH_IPV6_ONLY	0x0100
+#define SEARCH_PROTO		0x0200
+#define SEARCH_UNIX		0x0400
 
-#define PROTOCOL_TCP    3
-#define PROTOCOL_UDP    2
-#define PROTOCOL_RAW    1
+#define PROTOCOL_UNIX		7
+#define PROTOCOL_MAX_V6 	6
+#define PROTOCOL_TCP6   	6
+#define PROTOCOL_UDP6   	5
+#define PROTOCOL_RAW6   	4
+#define PROTOCOL_MAX_V4 	3
+#define PROTOCOL_TCP		3
+#define PROTOCOL_UDP    	2
+#define PROTOCOL_RAW    	1
+#define PROTOCOL_NOTHING	0
 
 #define INT_BIT	    (sizeof(int) * CHAR_BIT)
 #define SET_PORT(p) do { ports[p / INT_BIT] |= 1 << (p % INT_BIT); } while (0)
@@ -65,8 +78,12 @@
 typedef struct {
 	ino_t inode;
 	int fd;
+	int refcount, proto, flags, type;
+	char path[1024];
 	struct in_addr local_addr;
 	struct in_addr remote_addr;
+	struct in6_addr local6_addr;
+	struct in6_addr remote6_addr;
 	unsigned int local_port;
 	unsigned int remote_port;
 	unsigned char status, protocol;
@@ -81,9 +98,12 @@ const char *states[] = {
 uid_t o_uid;
 gid_t o_gid;
 pid_t o_pid;
+unsigned char o_protocol;
 char buf[1024], o_pname[8];
 DIR *proc, *fd;
 FILE *tcp, *udp, *raw;
+FILE *tcp6, *udp6, *raw6;
+FILE *funix;
 procnet_entry_t *netdata;
 unsigned int o_search = SEARCH_ALL;
 
@@ -124,7 +144,8 @@ void *xrealloc(void *ptr, size_t sz)
 
 void usage(const char *progname)
 {
-	fprintf(stderr, "usage: %s [-clh] [-p ports] [-U uid|user] [-G gid|group] [-P pid|process]\n", progname);
+	fprintf(stderr, "usage: %s [-46clhu] [-p ports] [-U uid|user] [-G gid|group] [-P pid|process] [-R protocol]\n", progname);
+	fprintf(stderr, "       protocol = 'tcp' or 'udp' or 'raw' or 'unix'\n");
 	exit(1);
 }
 
@@ -151,14 +172,24 @@ int digittoint(const char i)
 }
 
 /* lame function to switch between buffers easily. */
-int read_tcp_udp_raw(char *buf, int bufsize)
+int read_tcp_udp_raw(char *buf, int bufsize, char initialFileToRead)
 {
-	static char fc = PROTOCOL_TCP;
+	static char fc = PROTOCOL_NOTHING;
 	FILE *fileptr;
 
+	if (fc==PROTOCOL_NOTHING) fc=initialFileToRead;
       change:
 	switch (fc)
 	{
+	  case PROTOCOL_TCP6:
+		  fileptr = tcp6;
+		  break;
+	  case PROTOCOL_UDP6:
+		  fileptr = udp6;
+		  break;
+	  case PROTOCOL_RAW6:
+		  fileptr = raw6;
+		  break;
 	  case PROTOCOL_TCP:
 		  fileptr = tcp;
 		  break;
@@ -168,6 +199,9 @@ int read_tcp_udp_raw(char *buf, int bufsize)
 	  case PROTOCOL_RAW:
 		  fileptr = raw;
 		  break;
+	  case PROTOCOL_UNIX:
+		  fileptr = funix;
+		  break;
 	  case 0:
 		  return 0;
 	  default:
@@ -205,13 +239,32 @@ const char *get_program_name(pid_t pid)
 	return "unknown";
 }
 
+const unsigned char string_to_protocol(char *proto)
+{
+	if (strncmp(proto,"udp",3)==0) return PROTOCOL_UDP;
+	if (strncmp(proto,"tcp",3)==0) return PROTOCOL_TCP;
+	if (strncmp(proto,"raw",3)==0) return PROTOCOL_RAW;
+	if (strncmp(proto,"unix",4)==0) return PROTOCOL_UNIX;
+
+	return 0;
+}
+
 const char *protocol_to_string(unsigned int proto)
 {
-	const char *type[] = { "raw", "udp4", "tcp4", NULL };
+	const char *type[] = { "raw", "udp4", "tcp4", "raw6", "udp6", "tcp6", "unix", NULL };
 
 	return type[proto - 1];
 }
 
+const char *conn6_addr(struct in6_addr addr)
+{
+	char buf[1024];
+
+        inet_ntop( AF_INET6, &addr, buf, 1024);
+
+	return strdup(buf);
+}
+
 const char *conn_addr(struct in_addr addr)
 {
 	if (inet_lnaof(addr) == INADDR_ANY)
@@ -220,6 +273,23 @@ const char *conn_addr(struct in_addr addr)
 	return inet_ntoa(addr);
 }
 
+char *conn6_to_string(struct in6_addr addr, unsigned int port)
+{
+	char buf[1024];
+	int i = 0;
+
+	i += snprintf(buf, sizeof(buf), "%s:", conn6_addr(addr));
+	if (port != 0)
+		snprintf(buf + i, sizeof(buf) - i, "%u", port);
+	else
+	{
+		buf[i] = '*';
+		buf[i + 1] = '\0';
+	}
+
+	return strdup(buf);
+}
+
 char *conn_to_string(struct in_addr addr, unsigned int port)
 {
 	char buf[1024];
@@ -241,9 +311,24 @@ void display_record(procnet_entry_t *record, pid_t pid, const char *pname)
 {
 	char *sbuf, *sbuf2;
 	struct passwd *pwd;
+	int printStatus=1;
 
-	sbuf = conn_to_string(record->local_addr, record->local_port);
-	sbuf2 = conn_to_string(record->remote_addr, record->remote_port);
+	if (record->protocol<=PROTOCOL_MAX_V4) {
+		/* we have a IPv4 record */
+		sbuf = conn_to_string(record->local_addr, record->local_port);
+		sbuf2 = conn_to_string(record->remote_addr, record->remote_port);
+	} else {
+		if ((record->protocol<=PROTOCOL_MAX_V6) && (record->protocol>PROTOCOL_MAX_V4)) {
+			/* we have a IPv6 record */
+			sbuf = conn6_to_string(record->local6_addr, record->local_port);
+			sbuf2 = conn6_to_string(record->remote6_addr, record->remote_port);
+		} else {
+			/* we have a unix record */
+			sbuf = strdup(record->path);
+			sbuf2= strdup(" ");
+			printStatus=0;
+		}
+	}
 
 	pwd = getpwuid(record->uid);
 	if (pwd == NULL) {
@@ -251,20 +336,20 @@ void display_record(procnet_entry_t *record, pid_t pid, const char *pname)
 		printf("%-8s %-20s %-8u %-6s %-25s %-25s %s\n",
 			"N/A", pname, pid, protocol_to_string(record->protocol),
 			sbuf, sbuf2,
-			states[record->status - 1]);
+			(printStatus==1)?states[record->status - 1]:" ");
 	} else {
 		pwd->pw_name[8] = '\0';
 		printf("%-8s %-20s %-8u %-6s %-25s %-25s %s\n",
 			pwd->pw_name, pname, pid, protocol_to_string(record->protocol),
 			sbuf, sbuf2,
-			states[record->status - 1]);
+			(printStatus==1)?states[record->status - 1]:" ");
 	}
 
 	free(sbuf);
 	free(sbuf2);
 }
 
-unsigned int read_proc_net(void)
+unsigned int read_proc_net(char initialFileToRead)
 {
 	int d;
 	unsigned int i = 0, size = 256, total;
@@ -272,13 +357,20 @@ unsigned int read_proc_net(void)
 
 	netdata = xcalloc(sizeof(procnet_entry_t), size);
 
-	while ((protocol = read_tcp_udp_raw(buf, sizeof(buf))) != 0)
+	while ((protocol = read_tcp_udp_raw(buf, sizeof(buf), initialFileToRead)) != 0)
 	{
 		char *q, *x, *y;
 		int lport, rport;
 		uid_t uid;
 		ino_t inode;
 		unsigned char status;
+		/* for unix records */
+		int refcount;
+		int proto;
+		int flags;
+		int type;
+		char *path[1024];
+		int count;
 
 		if (i == size)
 		{
@@ -286,21 +378,90 @@ unsigned int read_proc_net(void)
 			netdata = xrealloc(netdata, (sizeof(procnet_entry_t) * size));
 		}
 
-		if (sscanf(buf, "%*d: %lX:%x %lX:%x %hx %*X:%*X %*X:%*X %*x %u %*u %u",
-			   (u_long *) &netdata[i].local_addr, &lport,
-			   (u_long *) &netdata[i].remote_addr, &rport, &status, 
-			   &uid, &inode) != 7)
-			continue;
-
-		netdata[i].local_port = lport;
-		netdata[i].remote_port = rport;
-		netdata[i].uid = uid;
-		netdata[i].inode = inode;
-		netdata[i].status = status;
-		netdata[i++].protocol = protocol;
+		if (protocol <= PROTOCOL_MAX_V4) {
+			/* we have an V4 entry */
+			if (sscanf(buf, "%*d: %lX:%x %lX:%x %hx %*X:%*X %*X:%*X %*x %u %*u %u",
+				   (u_long *) &netdata[i].local_addr, &lport,
+				   (u_long *) &netdata[i].remote_addr, &rport, &status, 
+				   &uid, &inode) != 7)
+				continue;
+			netdata[i].local_port = lport;
+			netdata[i].remote_port = rport;
+			netdata[i].uid = uid;
+			netdata[i].inode = inode;
+			netdata[i].status = status;
+			netdata[i++].protocol = protocol;
+		} else {
+			if ((protocol > PROTOCOL_MAX_V4) && (protocol <= PROTOCOL_MAX_V6)) {
+				/* we have an V6 entry */
+				if (sscanf(buf, "%*d: %2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx:%x %2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx:%x %hx %*X:%*X %*X:%*X %*x %u %*u %u",
+				   &netdata[i].local6_addr.s6_addr[ 3],
+				   &netdata[i].local6_addr.s6_addr[ 2],
+				   &netdata[i].local6_addr.s6_addr[ 1],
+				   &netdata[i].local6_addr.s6_addr[ 0],
+				   &netdata[i].local6_addr.s6_addr[ 7],
+				   &netdata[i].local6_addr.s6_addr[ 6],
+				   &netdata[i].local6_addr.s6_addr[ 5],
+				   &netdata[i].local6_addr.s6_addr[ 4],
+				   &netdata[i].local6_addr.s6_addr[11],
+				   &netdata[i].local6_addr.s6_addr[10],
+				   &netdata[i].local6_addr.s6_addr[ 9],
+				   &netdata[i].local6_addr.s6_addr[ 8],
+				   &netdata[i].local6_addr.s6_addr[15],
+				   &netdata[i].local6_addr.s6_addr[14],
+				   &netdata[i].local6_addr.s6_addr[13],
+				   &netdata[i].local6_addr.s6_addr[12],
+				   &lport,
+				   &netdata[i].remote6_addr.s6_addr[ 3],
+				   &netdata[i].remote6_addr.s6_addr[ 2],
+				   &netdata[i].remote6_addr.s6_addr[ 1],
+				   &netdata[i].remote6_addr.s6_addr[ 0],
+				   &netdata[i].remote6_addr.s6_addr[ 7],
+				   &netdata[i].remote6_addr.s6_addr[ 6],
+				   &netdata[i].remote6_addr.s6_addr[ 5],
+				   &netdata[i].remote6_addr.s6_addr[ 4],
+				   &netdata[i].remote6_addr.s6_addr[11],
+				   &netdata[i].remote6_addr.s6_addr[10],
+				   &netdata[i].remote6_addr.s6_addr[ 9],
+				   &netdata[i].remote6_addr.s6_addr[ 8],
+				   &netdata[i].remote6_addr.s6_addr[15],
+				   &netdata[i].remote6_addr.s6_addr[14],
+				   &netdata[i].remote6_addr.s6_addr[13],
+				   &netdata[i].remote6_addr.s6_addr[12],
+				   &rport, &status, 
+				   &uid, &inode) != 37)
+					continue;
+				netdata[i].local_port = lport;
+				netdata[i].remote_port = rport;
+				netdata[i].uid = uid;
+				netdata[i].inode = inode;
+				netdata[i].status = status;
+				netdata[i++].protocol = protocol;
+			} else {
+				/* we have an unix entry */
+				count=sscanf(buf, "%*x: %u %u %u %u %u %u %s",
+					&refcount,
+					&proto,
+					&flags,
+					&type,
+					&status,
+					&inode,
+					&path);
+				if ((count!=6) && (count!=7)) continue;
+				snprintf(netdata[i].path,1024,"%s",path);
+				netdata[i].refcount=refcount;
+				netdata[i].proto=proto;
+				netdata[i].flags=flags;
+				netdata[i].type=type;
+				netdata[i].inode = inode;
+				netdata[i].status = status;
+				netdata[i++].protocol = protocol;
+			}
+		}	
 	}
 
 	total = i;
+
 	qsort(netdata, total, sizeof(procnet_entry_t), compare);
 
 	return total;
@@ -361,15 +522,29 @@ int main(int argc, char *argv[])
 	struct group *grp;
 	struct dirent *procent, *fdent;
 	int ch, i;
+	char initialFileToRead=PROTOCOL_MAX_V6;
 	unsigned int total;
 
-	while ((ch = getopt(argc, argv, "clp:G:U:P:h")) != EOF)
+	while ((ch = getopt(argc, argv, "46clp:G:U:P:hR:u")) != EOF)
 		switch (ch)
 		{
+		  case '4':
+			  o_search |= SEARCH_IPV4_ONLY;
+			  break;
+		  case '6':
+			  o_search |= SEARCH_IPV6_ONLY;
+			  break;
+		  case 'u':
+			  initialFileToRead=PROTOCOL_UNIX;
+			  break;
 		  case 'p':
 			  o_search |= SEARCH_PORTS;
 			  parse_ports(optarg);
 			  break;
+		  case 'R':
+			  o_search |= SEARCH_PROTO;
+			  o_protocol=string_to_protocol(optarg);
+			  break;
 		  case 'c':
 			  o_search |= SEARCH_CONN;
 			  break;
@@ -416,14 +591,39 @@ int main(int argc, char *argv[])
 	if ((raw = fopen("/proc/net/raw", "r")) == NULL)
 		handle_missing_file("/proc/net/raw");
 
+	if ((tcp6 = fopen("/proc/net/tcp6", "r")) == NULL)
+		handle_missing_file("/proc/net/tcp6");
+
+	if ((udp6 = fopen("/proc/net/udp6", "r")) == NULL)
+		handle_missing_file("/proc/net/udp6");
+
+	if ((raw6 = fopen("/proc/net/raw6", "r")) == NULL)
+		handle_missing_file("/proc/net/raw6");
+
+	if ((funix = fopen("/proc/net/unix", "r")) == NULL) {
+		/* 
+		 * independent of -u, we need do avoid PROTOCOL_UNIX if this file 
+		 * is not available
+		 */
+		initialFileToRead=PROTOCOL_MAX_V6;
+	}
+
 	if ((proc = opendir("/proc")) == NULL)
 		abort();
 
-	total = read_proc_net();
+	total = read_proc_net(initialFileToRead);
 
+	/*
+	 * program has been abort()ed if file not available
+	 * except funix
+	 */
 	fclose(tcp);
 	fclose(udp);
 	fclose(raw);
+	fclose(tcp6);
+	fclose(udp6);
+	fclose(raw6);
+	if (funix) fclose(funix);
 
 	printf("%-8s %-20s %-8s %-6s %-25s %-25s %s\n", "USER", "PROCESS", "PID", "PROTO", "SOURCE ADDRESS", "FOREIGN ADDRESS", "STATE");
 
@@ -456,6 +656,7 @@ int main(int argc, char *argv[])
 				int display = 0;
 				pid_t pid = atoi(procent->d_name);
 
+
 				pn = get_program_name(pid);
 
 				if (o_search == 0)
@@ -464,6 +665,39 @@ int main(int argc, char *argv[])
 					continue;
 				}
 
+				if (o_search & SEARCH_PROTO)
+				{
+					if (o_protocol == PROTOCOL_UDP) {
+						if ((ptr->protocol == PROTOCOL_UDP) ||
+						    (ptr->protocol == PROTOCOL_UDP6))
+							display = 1;
+					}
+					if (o_protocol == PROTOCOL_TCP) {
+						if ((ptr->protocol == PROTOCOL_TCP) ||
+						    (ptr->protocol == PROTOCOL_TCP6))
+							display = 1;
+					}
+					if (o_protocol == PROTOCOL_RAW) {
+						if ((ptr->protocol == PROTOCOL_RAW) ||
+						    (ptr->protocol == PROTOCOL_RAW6))
+							display = 1;
+					}
+					if (o_protocol == PROTOCOL_UNIX) {
+						if (ptr->protocol == PROTOCOL_UNIX)
+							display = 1;
+					}
+				}
+				if (o_search & SEARCH_IPV4_ONLY)
+				{
+					if (ptr->protocol<=PROTOCOL_MAX_V4)
+						display = 1;
+				}
+				if (o_search & SEARCH_IPV6_ONLY)
+				{
+					if ((ptr->protocol<=PROTOCOL_MAX_V6)
+					 && (ptr->protocol>PROTOCOL_MAX_V4))
+						display = 1;
+				}
 				if (o_search & SEARCH_PID)
 				{
 					if (o_pid == atoi(procent->d_name))
